import re
import warnings

from edwardsserial.serial_protocol import SerialProtocol


class Gauge(SerialProtocol):
    GAUGE_NAME = re.compile(
        "[A-Z0-9]{0,4}$"
    )  # [a-z] does not get an error response but does not change the name.
    UNITS = {
        59: "Pa",
        66: "V",
        81: "%",
    }

    GAUGE_TYPE = {
        0: "Unknown Device",
        1: "No Device",
        2: "EXP_CM",
        3: "EXP_STD",
        4: "CMAN_S",
        5: "CMAN_D",
        6: "TURBO",
        7: "APGM",
        8: "APGL",
        9: "APGXM",
        10: "APGXH",
        11: "APGXL",
        12: "ATCA",
        13: "ATCD",
        14: "ATCM",
        15: "WRG",
        16: "AIMC",
        17: "AIMN",
        18: "AIMS",
        19: "AIMX",
        20: "AIGC_I2R",
        21: "AIGX_2FIL",
        22: "ION_EB",
        23: "AIGXS",
        24: "USER",
        25: "ASG",
    }

    GAUGE_STATE = {
        0: "Gauge Not Connected",
        1: "Gauge Connected",
        2: "New Gauge ID",
        3: "Gauge Change",
        4: "Gauge In ALert",
        5: "Off",
        6: "Striking",
        7: "Initialising",
        8: "Calibrate",
        9: "Zeroing",
        10: "Degassing",
        11: "On",
        12: "Inhibited",
    }

    GAS_TYPE = {
        0: "Nitrogen",
        1: "Helium",
        2: "Argon",
        3: "Carbon Dioxide",
        4: "Neon",
        5: "Krypton",
        6: "Voltage",
    }

    ASG_RANGE = {
        0: "1000 mbar",
        1: "2000 mbar",
    }

    def __init__(self, port, object_id):
        super().__init__(port)
        self.OBJECT_ID = object_id

    @property
    def pressure(self):
        value, unit, state = self._check_alert(self.OBJECT_ID)
        if int(state) not in [8, 11]:
            return None
        unit = int(unit)
        if unit == 81:
            return int(value)
        return float(value)

    @property
    def unit(self):
        value, unit, state = self._check_alert(self.OBJECT_ID)
        return self.UNITS.get(int(unit))

    @property
    def state(self):
        value, unit, state = self._check_alert(self.OBJECT_ID)
        return f"{state}: {self.GAUGE_STATE.get(int(state))}"

    @property
    def type(self):
        config_type, gauge_type = self.send_message("?S", self.OBJECT_ID, 5)
        return f"{gauge_type}: {self.GAUGE_TYPE.get(int(gauge_type))}"

    @property
    def name(self):
        config_type, name = self.send_message("?S", self.OBJECT_ID, 68)
        return name

    @name.setter
    def name(self, value):
        if not self.GAUGE_NAME.match(value):
            raise ValueError(
                "Wrong name format: Only 4 characters of [A-Z0-9] are allowed."
            )
        self.send_message("!S", self.OBJECT_ID, f"68;{value}")

    @property
    def gas_type(self):
        config_type, gas_type, gas_filter = self.send_message("?S", self.OBJECT_ID, 7)
        return f"{gas_type}: {self.GAS_TYPE.get(int(gas_type))}"

    @gas_type.setter
    def gas_type(self, value):
        if value not in self.GAS_TYPE.keys():
            raise ValueError(f"Value must be a key from {self.GAS_TYPE}")
        self.send_message("!S", self.OBJECT_ID, f"7;{value};{int(self.filter)}")

    @property
    def filter(self):
        """ Moving average filter for the pressure of 1 second.

        Returns
        -------
        Filter active: True or False
        """
        config_type, gas_type, gas_filter = self.send_message("?S", self.OBJECT_ID, 7)
        return bool(int(gas_filter))

    @filter.setter
    def filter(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError(f"Value must be boolean.")
        self.send_message("!S", self.OBJECT_ID, f"7;{self.gas_type[0]};{int(value)}")

    @property
    def ASG_range(self):
        config_type, asg_range = self.send_message("!S", self.OBJECT_ID, 6)
        return f"{int(asg_range)}: {self.ASG_RANGE.get(int(asg_range))}"

    @ASG_range.setter
    def ASG_range(self, value):
        if value not in self.ASG_RANGE.keys():
            raise ValueError(f"Value must be a key from {self.ASG_RANGE}")
        self.send_message("!S", self.OBJECT_ID, f"{6};{value}")

    def on(self):
        self.send_message("!C", self.OBJECT_ID, 1)

    def off(self):
        self.send_message("!C", self.OBJECT_ID, 0)

    def zero(self):
        self.send_message("!C", self.OBJECT_ID, 3)

    def calibrate(self):
        self.send_message("!C", self.OBJECT_ID, 4)

    def degas(self):
        self.send_message("!C", self.OBJECT_ID, 5)

    def new_id(self):
        self.send_message("!C", self.OBJECT_ID, 2)



import logging
import re
from warnings import warn

import serial

log = logging.getLogger(__name__)


class AlertID(Warning):
    ALERT_ID = {
        0: "No Alert",
        1: "ADC Fault",
        2: "ADC Not Ready",
        3: "Over Range",
        4: "Under Range",
        5: "ADC Invalid",
        6: "No Gauge",
        7: "Unknown",
        8: "Not Supported",
        9: "New ID",
        10: "Over Range",
        11: "Under Range",
        12: "Over Range",
        13: "Ion Em Timeout",
        14: "Not Struck",
        15: "Filament Fail",
        16: "Mag Fail",
        17: "Striker Fail",
        18: "Not Struck",
        19: "Filament Fail",
        20: "Cal Error",
        21: "Initialising",
        22: "Emission Error",
        23: "Over Pressure",
        24: "ASG Cant Zero",
        25: "RampUp Timeout",
        26: "Droop Timeout",
        27: "Run Hours High",
        28: "SC Interlock",
        29: "ID Volts Error",
        30: "Serial ID Fail",
        31: "Upload Active",
        32: "DX Fault",
        33: "Temp Alert",
        34: "SYSI Inhibit",
        35: "Ext Inhibit",
        36: "Temp Inhibit",
        37: "No Reading",
        38: "No Message",
        39: "NOV Failure",
        40: "Upload Timeout",
        41: "Download Failed",
        42: "No Tube",
        43: "Use Gauges 4-6",
        44: "Degas Inhibited",
        45: "IGC Inhibited",
        46: "Brownout/Short",
        47: "Service due",
    }

    def __init__(self, alert_id):
        message = f"{alert_id} ({self.ALERT_ID.get(alert_id)})"
        super().__init__(message)
        # self.id = alert_id


class ErrorResponse(Exception):

    ERROR_CODES = {
        1: "Invalid command for object ID",
        2: "Invalid query/command",
        3: "Missing parameter",
        4: "Parameter out of range",
        5: "Invalid command in current state - e.g. serial command to start or stop when in parallel control mode",
        6: "Data checksum error",
        7: "EEPROM read or write error",
        8: "Operation took too long",
        9: "Invalid config ID",
    }

    def __init__(self, error_code):
        message = f"Device responded with error code {error_code}: {self.ERROR_CODES.get(int(error_code))}"
        super().__init__(message)


class SerialProtocol:
    BAUDRATE = 9600
    COMMAND = re.compile(r"![CS]\d{1,5} (.+;?)+\r", flags=re.ASCII)
    SETUP_QUERY = re.compile(r"\?[S]\d{1,5}( \d+)*\r", flags=re.ASCII)
    VALUE_QUERY = re.compile(r"\?[V]\d{1,5}\r", flags=re.ASCII)
    MESSAGE = re.compile(
        f"{COMMAND.pattern}|{SETUP_QUERY.pattern}|{VALUE_QUERY.pattern}"
    )

    ERROR_RESPONSE = re.compile(
        r"\*[CSV]\d{1,5} (?P<error_code>\d{1,2})\r", flags=re.ASCII
    )
    DATA_RESPONSE = re.compile(r"=[SV]\d{1,5} (?P<data>.+;?)+\r", flags=re.ASCII)
    RESPONSE = re.compile(f"{ERROR_RESPONSE.pattern}|{DATA_RESPONSE.pattern}")

    STATE = {
        0: "Off State",
        1: "Off Goiing On State",
        2: "On Going Off Shutdown State",
        3: "On Going Off Normal State",
        4: "On State",
    }

    def __init__(self, port: str):
        self.port = port

    def _check_alert(self, object_id):
        *values, alert_id, priority = self.send_message("?V", object_id)
        alert_id = int(alert_id)
        if alert_id:
            warn(AlertID(alert_id))
        return values

    @classmethod
    def _create_message(cls, operation: str, object_id: int, data=None) -> str:
        data = f" {data}" if data is not None else ""
        message = f"{operation}{object_id}{data}\r"
        if not cls.MESSAGE.match(message):
            raise ValueError(
                f"Serial message {message.encode('unicode-escape')!r} does not have the correct format of {cls.MESSAGE}."
            )
        return message

    def send_message(self, operation, object_id, data=None):
        message = self._create_message(operation, object_id, data=data)
        with serial.Serial(self.port, timeout=1, baudrate=self.BAUDRATE) as ser:
            ser.write(message.encode("ascii"))
            response = ser.read_until(b"\r").decode("ascii")
        log.debug(f"send_message: response={response}")
        response = self.RESPONSE.match(response)
        if not response:
            raise ConnectionError("No serial connection to device.")
        groups = response.groupdict()
        error_code = groups.get("error_code")
        if error_code:
            error_code = int(error_code)
            if error_code:
                raise ErrorResponse(error_code=error_code)
            return
        return groups.get("data").split(";")



from abc import ABC, abstractmethod

from edwardsserial.serial_protocol import SerialProtocol


class Pump(SerialProtocol, ABC):
    PUMP_ID: int
    SPEED_ID: int
    POWER_ID: int

    PUMP_STATE = {
        0: "Stopped",
        1: "Starting Delay",
        2: "Stopping Short Delay",
        3: "Stopping Normal Delay",
        4: "Running",
        5: "Accelerating",
        6: "Fault Braking",
        7: "Braking",
    }
    PUMP_TYPE = {
        0: "No Pump",
        1: "EXDC",
        3: "EXT75DX",
        4: "EXT255DC",
        8: "Mains Backing Pump",
        9: "Serial Pump",
        10: "nEXT - 485",
        11: "nEXT - 232",
        12: "nXDS",
        # 16: "nEXT",  # todo: check if this is True
        99: "Not yet identified",
    }

    def on(self):
        self.send_message("!C", self.PUMP_ID, 1)

    def off(self):
        self.send_message("!C", self.PUMP_ID, 0)

    @property
    def state(self):
        state = int(*self._check_alert(self.PUMP_ID))
        return f"{state}: {self.PUMP_STATE.get(state)}"

    @property
    def speed(self):
        return float(*self._check_alert(self.SPEED_ID))

    @property
    def power(self):
        return float(*self._check_alert(self.POWER_ID))

    @property
    def type(self):
        config_type, pump_type = self.send_message("?S", self.PUMP_ID, 3)
        return f"{pump_type}: {self.PUMP_TYPE.get(int(pump_type))}"


class TurboPump(Pump):
    PUMP_ID = 904
    SPEED_ID = 905
    POWER_ID = 906
    NORMAL_ID = 907
    STANDBY_ID = 908
    CYCLE_ID = 909

    @property
    def normal(self):
        value = int(*self._check_alert(self.NORMAL_ID))
        if value == 0:
            return False
        if value == 4:
            return True
        raise ValueError(f"Got state={value}. Expected 0 or 4.")

    @property
    def standby(self):
        value = int(*self._check_alert(self.STANDBY_ID))
        if value == 0:
            return False
        if value == 4:
            return True
        raise ValueError(f"Got state={value}. Expected 0 or 4.")

    @standby.setter
    def standby(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("Value must be of type bool.")
        self.send_message("?C", self.STANDBY_ID, int(value))

    @property
    def cycle_time(self):
        value, state = self._check_alert(self.CYCLE_ID)
        return int(value)

    @property
    def delay(self):
        config_type, delay = self.send_message("?S", self.PUMP_ID, 21)
        return int(delay)

    @delay.setter
    def delay(self, value: int):
        if value not in range(0, 100):
            raise ValueError("Must be between 0 and 99")
        self.send_message("!S", self.PUMP_ID, f"21;{value}")


class BackingPump(Pump):
    PUMP_ID = 910
    SPEED_ID = 911
    POWER_ID = 912

    @property
    def sequence_options(self):
        pass



from edwardsserial.serial_protocol import SerialProtocol


# make this as a descriptor that calls the owners pump start methods if available,
# since TIC does not forward some commands
class nEXT(SerialProtocol):
    def __get__(self, instance, owner):
        pass

    def __set__(self, instance, value):
        pass



from edwardsserial.serial_protocol import SerialProtocol
from edwardsserial.tic.gauge import Gauge
from edwardsserial.tic.pump import BackingPump, TurboPump


class TIC(SerialProtocol):
    def __init__(self, port):
        super().__init__(port)
        self._turbo_pump = TurboPump(port)
        self._backing_pump = BackingPump(port)
        self._gauge1 = Gauge(port, 913)
        self._gauge2 = Gauge(port, 914)
        self._gauge3 = Gauge(port, 915)

    PRESSURE_UNITS = {
        1: "kPa",
        2: "mbar",
        3: "Torr",
    }

    @property
    def turbo_pump(self):
        return self._turbo_pump

    @property
    def backing_pump(self):
        return self._backing_pump

    @property
    def gauge1(self):
        return self._gauge1

    @property
    def gauge2(self):
        return self._gauge2

    @property
    def gauge3(self):
        return self._gauge3

    @property
    def gauge_values(self):
        answer = self.send_message("?V", 940)
        values = {
            int(gauge): float(value) for gauge, value in zip(answer[::2], answer[1::2])
        }
        return values

    @property
    def status(self):
        state = self._check_alert(933)
        return f"{state}: {self.STATE[state]}"

    @property
    def pressure_units(self):
        unit = self.send_message("?S", 929)
        return f"{unit}: {self.PRESSURE_UNITS.get(unit)}"

    @pressure_units.setter
    def pressure_units(self, value):
        if value not in self.PRESSURE_UNITS.keys():
            raise ValueError(f"Value must be a key from {self.PRESSURE_UNITS}")
        self.send_message("!S", 929, value)

    @property
    def display_contrast(self):
        return int(self.send_message("?S", 925))

    @display_contrast.setter
    def display_contrast(self, value):
        if value not in range(-5, 16):
            raise ValueError(f"Value must be in {range(-5,16)}")
        self.send_message("!S", 925, value)

    @property
    def internal_temperature(self):
        temperature = self._check_alert(920)
        return float(temperature)

    @property
    def power_supply_temperature(self):
        temperature = self._check_alert(919)
        return float(temperature)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edwardsserial', 'edwardsserial.tic']

package_data = \
{'': ['*']}

install_requires = \
['pyserial']

setup_kwargs = {
    'name': 'edwardsserial',
    'version': '0.3.1',
    'description': 'Python API implementing the serial protocoll from edwards vacuum.',
    'long_description': '# TIC\n\nPython object-oriented wrapper for the TIC Turbo- and Instrument Controller from Edwards (https://shop.edwardsvacuum.com/products/d39722000/view.aspx).',
    'author': 'Jan Petermann',
    'author_email': 'jpeterma@physnet.uni-hamburg.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/codingcoffeebean/edwardsserial',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
